package org.agil.core.spatialindex;

/*
 *  A HyperCube in the n-dimensional space. It is represented by two points of n dimensions each.
 *  <p>
 *  Implements basic calculation functions, like <B>getArea()</B> and <B>getUnionMbb()</B>.
 *  <p>
 *  Created: Tue May 18 14:33:37 1999
 *  <p>
 *  @author Marios Hadjieleftheriou
 *  @version 2
 */
/**
 *  HyperCube de n dimensiones. Se representa por dos puntos de n dimensiones
 *  cada uno. Implementa las funciones de calculo basicas, <b>como getArea()
 *  </b> y <b>getUnionMbb()</b> .
 *
 *@author    Chris Seguin
 */
public class HyperCube implements Cloneable {
	private Point low;
	private Point high;

	private static final float EPSILON = 0.0001f;


	/**
	 *  Crea un nuevo HyperCube a partir de dos puntos
	 *
	 *@param  p1  Punto Inferior-Izquierda
	 *@param  p2  Punto Superior-Derecha
	 */
	public HyperCube(Point p1, Point p2) {
		if (p1 == null || p2 == null) {
			throw new IllegalArgumentException("Points cannot be null.");
		}
		if (p1.getDimension() != p2.getDimension()) {
			throw new IllegalArgumentException("Points must be of the same dimension.");
		}
		for (int i = 0; i < p1.getDimension(); i++) {
			if (p1.getFloatCoordinate(i) > p2.getFloatCoordinate(i)) {
				throw new IllegalArgumentException("Give lower left corner first and upper right corner afterwards.");
			}
		}
		this.low = (Point) p1.clone();
		this.high = (Point) p2.clone();
	}


	/**
	 *@return    la dimension de los puntos
	 */
	public int getDimension() {
		return low.getDimension();
	}


	/**
	 *@return    el punto Inferior-Izquierda
	 */
	public Point getLowPoint() {
		return (Point) low.clone();
	}


	/**
	 *@return    el punto Superior-Derecha
	 */
	public Point getHighPoint() {
		return (Point) high.clone();
	}


	/**
	 *  Retorna el area de este HyperCube.
	 *
	 *@return    el area.
	 */
	public float getArea() {
		float area = 1;
		for (int i = 0; i < getDimension(); i++) {
			area *= high.getFloatCoordinate(i) - low.getFloatCoordinate(i);
		}
		return area;
	}


	/**
	 *  El criterio de MINIMA DISTANCIA descrito por Roussopoulos.
	 *
	 *@param  p  punto desde el cual calcular la distancia a este HyperCube.
	 *@return    The MinDist value
	 */
	public float getMinDist(Point p) {
		if (p == null) {
			throw new IllegalArgumentException("Point cannot be null.");
		}
		if (p.getDimension() != getDimension()) {
			throw new IllegalArgumentException("Point dimension is different from HyperCube dimension.");
		}
		float ret = 0;
		for (int i = 0; i < getDimension(); i++) {
			float q = p.getFloatCoordinate(i);
			float s = low.getFloatCoordinate(i);
			float t = high.getFloatCoordinate(i);
			float r;
			if (q < s) {
				r = s;
			}
			else if (q > t) {
				r = t;
			}
			else {
				r = q;
			}
			ret += Math.pow(Math.abs(q - r), 2);
		}
		return ret;
	}


	/**
	 *  Compara el HyperCube actual con otro.
	 *
	 *@param  h  HyperCube a comparar
	 *@return    true si los dos puntos del HyperCube actual son iguales a los de
	 *      <i>h</i>
	 */
	public boolean equals(HyperCube h) {
		if (low.equals(h.low) && high.equals(h.high)) {
			return true;
		}
		else {
			return false;
		}
	}


	/**
	 *  Comprueba si <b>h</b> tiene algun punto en comun con el HyperCube actual.
	 *  Si <b>h</b> esta dentro de este objeto o viceversa, retorna true.
	 *
	 *@param  h  Description of Parameter
	 *@return    true si <b>h</b> intersecta con este HyperCube.
	 */
	public boolean intersects(HyperCube h) {
		if (h == null) {
			throw new IllegalArgumentException("HyperCube cannot be null.");
		}
		if (h.getDimension() != getDimension()) {
			throw new IllegalArgumentException("HyperCubes have different dimensions.");
		}
		boolean intersect = true;
		for (int i = 0; i < getDimension(); i++) {
			if (mayor(low.getFloatCoordinate(i), h.high.getFloatCoordinate(i)) ||
					menor(high.getFloatCoordinate(i), h.low.getFloatCoordinate(i))) {
				intersect = false;
				break;
			}
		}
		return intersect;
	}


	/**
	 *  Comprueba si <b>h</b> esta contenido dentro de este HyperCube. Si <b>h
	 *  </b> es exactamente el mismo, se considera contenido.
	 *
	 *@param  h  Description of Parameter
	 *@return    true si <b>h</b> esta dentro.
	 */
	public boolean contains(HyperCube h) {
		if (h == null) {
			throw new IllegalArgumentException("HyperCube cannot be null.");
		}
		if (h.getDimension() != getDimension()) {
			throw new IllegalArgumentException("HyperCubes have different dimensions.");
		}
		boolean contained = true;
		for (int i = 0; i < getDimension(); i++) {
//			if (low.getFloatCoordinate(i) > h.low.getFloatCoordinate(i) || high.getFloatCoordinate(i) < h.high.getFloatCoordinate(i)) {
			if (mayor(low.getFloatCoordinate(i), h.low.getFloatCoordinate(i)) ||
					menor(high.getFloatCoordinate(i), h.high.getFloatCoordinate(i))) {
				contained = false;
				break;
			}
		}
		return contained;
	}


	/**
	 *  Comprueba si <b>p</b> esta contenido en este HyperCube. Si <b>p</b> es
	 *  uno de los puntos limites del Rectangulo definido por este HyperCube, se
	 *  considera contenido.
	 *
	 *@param  p  Description of Parameter
	 *@return    true <b>p</b> esta dentro.
	 */
	public boolean contains(Point p) {
		if (p == null) {
			throw new IllegalArgumentException("Point cannot be null.");
		}
		if (p.getDimension() != getDimension()) {
			throw new IllegalArgumentException("Point dimension is different from HyperCube dimension.");
		}
		return contains(new HyperCube(p, p));
	}


	/**
	 *  Retorna el area de la region intersectada entre los dos HyperCubes. Casos
	 *  Posibles: ------- ------- --------- --------- ------ ------ |2 | |2 | |2
	 *  | |1 | |2 | |1 | --|-- | | --|-- | ------ | | ------ | --|------|--
	 *  --|------|-- |1 | | | | |1 | | ||1 | | ||2 | | |1 | | | |2 | | | --|-- |
	 *  | --|-- | ------ | | ------ | --|------|-- --|------|-- ------- -------
	 *  --------- --------- ------ ------
	 *
	 *@param  h  el HyperCube dado.
	 *@return    Area de la region intersectada.
	 */
	public float intersectingArea(HyperCube h) {
		if (!intersects(h)) {
			return 0;
		}
		else {
			float ret = 1;
			for (int i = 0; i < getDimension(); i++) {
				float l1 = low.getFloatCoordinate(i);
				float h1 = high.getFloatCoordinate(i);
				float l2 = h.low.getFloatCoordinate(i);
				float h2 = h.high.getFloatCoordinate(i);
				if (l1 <= l2 && h1 <= h2) {
					// cube1 left of cube2.
					ret *= (h1 - l1) - (l2 - l1);
				}
				else if (l2 <= l1 && h2 <= h1) {
					// cube1 right of cube2.
					ret *= (h2 - l2) - (l1 - l2);
				}
				else if (l2 <= l1 && h1 <= h2) {
					// cube1 inside cube2 or cube1 crossed by cube2.
					ret *= h1 - l1;
				}
				else if (l1 <= l2 && h2 <= h1) {
					// cube1 contains cube2 or cube1 crosses cube2.
					ret *= h2 - l2;
				}
			}
			if (ret <= 0) {
				throw new ArithmeticException("Intersecting area cannot be negative!");
			}
			return ret;
		}
	}


	/**
	 *  Retorna un nuevo HyperCubo que representa la union de los dos.
	 *
	 *@param  h  el HyperCube dado
	 *@return    el HyperCube resultante de la union de los dos
	 */
	public HyperCube combinedMbb(HyperCube h) {
		if (h == null) {
			throw new IllegalArgumentException("HyperCube cannot be null.");
		}
		if (h.getDimension() != getDimension()) {
			throw new IllegalArgumentException("HyperCubes must be of the same dimension.");
		}
		float[] min = new float[getDimension()];
		float[] max = new float[getDimension()];
		for (int i = 0; i < getDimension(); i++) {
			min[i] = Math.min(low.getFloatCoordinate(i), h.low.getFloatCoordinate(i));
			max[i] = Math.max(high.getFloatCoordinate(i), h.high.getFloatCoordinate(i));
		}
		return new HyperCube(new Point(min), new Point(max));
	}


	/**
	 *  Creates an exact copy of this object.
	 *
	 *@return    Description of the Returned Value
	 */
	public Object clone() {
		return new HyperCube(low, high);
		//return this;
	}


	/**
	 *  Converts to a String representation of the object.
	 *
	 *@return    A string representation of the object.
	 */
	public String toString() {
		return low.toString() + ": " + high.toString();
	}


	/**
	 *  Description of the Method
	 *
	 *@param  f1  Description of Parameter
	 *@param  f2  Description of Parameter
	 *@return     Description of the Returned Value
	 */
	boolean mayor(float f1, float f2) {
		return ((f1 + EPSILON) > f2);
	}


	/**
	 *  Description of the Method
	 *
	 *@param  f1  Description of Parameter
	 *@param  f2  Description of Parameter
	 *@return     Description of the Returned Value
	 */
	boolean menor(float f1, float f2) {
		return ((f1 - EPSILON) < f2);
	}


	/**
	 *  Toma un array de HyperCubes y calcula el MBB de la union de todos ellos.
	 *
	 *@param  a  el array de HyperCubes. |return el MBB de su union.
	 *@return    Description of the Returned Value
	 */
	public static HyperCube combinedMbb(HyperCube[] a) {
		if (a == null || a.length == 0) {
			throw new IllegalArgumentException("HyperCube array is empty.");
		}
		HyperCube h = (HyperCube) a[0].clone();
		for (int i = 1; i < a.length; i++) {
			h = h.combinedMbb(a[i]);
		}
		return h;
	}

}
// HyperCube
